# This module contains wrapper functions around some Keras 3 functions
# to make them more easy and intuitive to use.
import keras
from keras import ops, random


def binomial(shape: tuple):
    """
    Samples from a binomial distribution of a given shape
    Args:
        shape: 

    Returns:

    """
    pass