# Copyright 2024 Khawaja Abaid Ullah
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keras.backend import backend


if backend() == "tensorflow":
    from teras._src.backend.tensorflow import *

elif backend() == "torch":
    from teras._src.backend.torch import *

elif backend() == "jax":
    from teras._src.backend.jax import *

else:
    raise ValueError("Backend not support. "
                     "Expected one of ['tensorflow', 'jax', 'torch']. "
                     f"Found {backend()}")
